#!/usr/bin/env python3
# SPDX-License-Identifier: MIT-0
#
# Implement the systemd notify protocol without external dependencies.
# Taken from the sd_notify man page

import bz2
import errno
import logging
import os
import shutil
import socket
from datetime import time
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path

def notify(message):
    if not message:
        raise ValueError("notify() requires a message")

    if not (socket_path := os.environ.get("NOTIFY_SOCKET")):
        return  # Not started by systemd

    if socket_path[0] not in ("/", "@"):
        raise OSError(errno.EAFNOSUPPORT, "Unsupported socket type")

    # Handle abstract socket.
    if socket_path[0] == "@":
        socket_path = "\0" + socket_path[1:]

    with socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM | socket.SOCK_CLOEXEC) as sock:
        sock.connect(socket_path)
        sock.sendall(message)

def notify_ready():
    """Tell systemd the service has started successfully"""
    notify(b"READY=1")


def file_and_stderr_logs(path: Path):
    """Make logging handlers for files & stderr"""
    path = path.expanduser()
    path.parent.mkdir(parents=True, exist_ok=True)

    # From Python logging cookbook: compress logs when they're rotated
    def rotator(source, dest):
        with open(source, 'rb') as f_in:
            with bz2.open(dest, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)
        os.remove(source)

    filehandler = TimedRotatingFileHandler(
        path,
        when='W1',  # W1 = Tuesday
        atTime=time(12, 0),  # At midday (after experiments finish)
        backupCount=40,  # weeks of logs to keep
    )
    filehandler.namer = lambda name: name + ".bz2"
    filehandler.rotator = rotator
    filehandler.setFormatter(logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - [%(filename)s:%(lineno)d] %(message)s'  # noqa
    ))

    # Also log to the journal (via stderr), which keeps its own timestamps
    streamhandler = logging.StreamHandler()
    streamhandler.setFormatter(logging.Formatter(
        '%(name)s - %(levelname)s - [%(filename)s:%(lineno)d] %(message)s'
    ))
    return [filehandler, streamhandler]
